/*****************************************************************************/
/*
                               [WASD]Query.c

A CGI-compliant script to search plain-text and HTML files on ODS-2 and ODS-5
volumes.  Extended file specifications may be expressed using either
RMS-escaped ("^_") or URL-escaped ("%nn") forbidden characters.  If a version
delimiter (';', with or without version number) is present in the path
specification then this script displays and anchors RMS-escaped and VMS syntax
file names.  If none is present it supplies URL-encoded file names.

Query works in concert with EXTRACT.C.

Accepts query-based searches (e.g. "/web/*.*?find+this+phrase") or form-based
fields briefly discussed below. By default searches the URL-supplied path, or
will use a path supplied with the "path=" query form field (via request
redirection to get the server to map the supplied path). A VMS directory
elipsis (i.e. "...") may be supplied in the path to result in a directory tree
search.

Hits in both plain-text files and HTML files provide a link to the extract
script. With plain text files the extract script extracts a section of the
file and presents it with some buttons to allow retrieving other sections or
all the document. When extracting HTML files it returns the entire document
but with each occurance of the hit enclosed by a '<span id="athitof1"></span>'
(yes, an oh not a zero) anchor that allows the specific hit to be jumped to
with relative document syntax.  The following tags do not have any content
included: <applet></applet>, <head></head>, <script></script>,
<server></server>, <style></style>, <title></title>.

Using a script-name prefixed path such as "/query/web/.../*.*" returns a simple
form for generating a search.

"Text" file extensions are predefined in the DEFAULT_TEXT_TYPES and
DEFAULT_HTML_TYPES macros.  To specify a custom list use /TEXT= and /HTML= or
to add other extensions to be considered text or HTML use /ADDTEXT= and
/ADDHTML= (not this is a comma-separated list with no extension period).  File
extensions may contain the asterisk wildcard character, representing zero or
more matching characters (e.g. "REPORT_*").


NOTE ON EXTENDED FILE SPECS
---------------------------
This script is ODS-5 volume compliant and will process extended file naming
(excluding those using Unicode characters).

It's design allows building on VAX VMS (which excludes ODS-5 support) and Alpha
VMS all versions.  When building on VAX all extended ODS code is "#ifdef"ed out
for efficiency reasons.  If built under an Alpha version that does not support
extended naming the ENAMEL.H header file provides an equivalent
(a build allowed by the bogus NAML created by ENAMEL.H) then it
can only process ODS-2 file names, however if built with VMS 7.2ff it will
process both, and the object module can still be linked and used on an earlier
version (although without ODS-5 capabilities of course).

This Alpha backward/forward compatibility is provided at runtime by checking
the version of VMS it is currently executing under.  If extended file
specification compliant then NAML structures are always used, otherwise NAMs. 
Hence, a VMS version that doesn't know about extended file specifications (and
doesn't have any ODS-5 volumes of course) never gets passed a NAML!


LOGICAL NAMES
-------------
QUERY$DBUG          turns on all "if (Debug)" statements
QUERY$PARAM         equivalent to (overrides) the command line
                    parameters/qualifiers (define as a system-wide logical)


HTML FORM ELEMENTS
------------------
case=                  case sensitive search (Y or N)
exact=                 exact number of records (for extract utility, Y or N)
extract=               number of line to pass to extract utility
hits=                  show all hits or document only (D or A)
html=                  comma-separated list of HTML file extensions
                       (overrides the /HTML and /ADDHTML qualifiers)
plain=                 if "yes" then treat HTML files as if plain-text
                       (i.e. search markup tags, everything!)
path=                  form supplied path (otherwise URL path)
search=                text to search for
target=                open new page in (e.g. "_blank", "_self")
text=                  comma-separated list of text file extensions
                       (overrides the /TEXT and /ADDTEXT qualifiers)
what=                  title on search results page

These could be used as in the following example (note that as this is in a
C-language comment "@" has been substituted for "*", they're just wildcards!):

<form action="/web/doc/.../@.@"> 
<input type="hidden" name="what" value="Example Search of /Web/ Documents">
<input type="submit" value="Search for:">
<input type="text" name="search" size="20">
<input type="reset" value="Reset">
<br>case sensitive?
<input type="radio" name="case" value="N" checked>N
<input type="radio" name="case" value="Y">Y
<br>extract this number of lines around a &quot;hit&quot;:
<input type="text" name="extract" value="20" size="3" maxlength="3">
</form>


QUALIFIERS
----------
/ABOUT=         synonym for /HELP
/ADDHTML=       additional list of comma separated HTML file types
/ADDTEXT=       additional list of comma separated TEXT file types
/CHARSET=       "Content-Type: text/html; charset=...", empty suppress charset
/DBUG           turns on all "if (Debug)" statements
/EXTRACT=       path to extract script
/HELP=          URL for help on searching
/HTML=          complete list of comma separated HTML file types
/[NO]ODS5       control extended file specification (basically for testing)
/STYLE=         URL for site CSS style sheet
/TEXT=          complete list of comma separated TEXT file types
/TIMEFMT=       strftime() format string for last-modified time


BUILD DETAILS
-------------
See BUILD_QUERY.COM procedure.


COPYRIGHT
---------
Copyright (C) 1996-2021 Mark G.Daniel

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.


VERSION HISTORY (update SOFTWAREVN as well!)
---------------
28-JUL-2020  MGD  v4.2.0, use "athitof" instead of "__hit__"
                          bugfix; style tag processing
28-JUN-2019  MGD  v4.1.0, allow scripts to contain anything but </script>
                          QueryHtmlEscape() allow HTML entities remain intact
04-OCT-2014  MGD  v4.0.0, a nod to the twenty-first century
10-MAY-2005  MGD  v3.2.9, SWS 2.0 ignore query string components supplied as
                          command-line parameters differently to CSWS 1.2/3
23-DEC-2003  MGD  v3.2.8, minor conditional mods to support IA64
15-AUG-2003  MGD  v3.2.7, bugfix; move fragment *after* query string
23-JUN-2003  MGD  v3.2.6, record size increased to maximum (32767 bytes),
                          ignore RMS$_WLK from sys$open() in line with RMS$_PRV
12-APR-2003  MGD  v3.2.5, link colour changed to 0000cc
15-AUG-2002  MGD  v3.2.4, GetParameters() mod for direct CSWS 1.2 support
01-JUL-2001  MGD  v3.2.3, add 'SkipParameters' for direct OSU support
19-MAY-2001  MGD  v3.2.2, remove limitation in SameFileType() preventing
                          searching of multiple file versions
25-JAN-2001  MGD  v3.2.1, use <BODY> to terminate <HEAD> processing
28-OCT-2000  MGD  v3.2.0, use CGILIB object module
02-MAR-2000  MGD  v3.1.2, bugfix;ed again:^( rework SameFileType()
28-FEB-2000  MGD  v3.1.1, bugfix; SameFileType() wildcard processing
15-FEB-2000  MGD  v3.1.0, allow wildcarded file types
18-JAN-2000  MGD  v3.0.0, support extended file specifications (ODS-5)
07-AUG-1999  MGD  v2.9.0, use more of the CGILIB functionality,
                          plain-text files described using file name
24-APR-1999  MGD  v2.8.0, use CGILIB.C,
                          standard CGI environment (e.g. Netscape FastTrack)
18-FEB-1999  MGD  v2.7.2, Search[Text/Html]File() handling of SS$_PRV
                          (files with protection violations now just ignored)
20-NOV-1998  MGD  v2.7.1, exclude certain content (e.g. <SCRIPT>...</SCRIPT)
07-NOV-1998  MGD  v2.7.0, anchor hits within an HTML document,
                          additional information against file name
03-OCT-1998  MGD  v2.6.0, remove dependency on MapUrl_Map(),
                          provide content-type "; charset=...",
                          accomodations for OSU environment,
                          more general maintenance
24-JUL-1998  MGD  v2.5.1, suppress table background colours if empty
20-MAY-1998  MGD  v2.5.0, general maintenance,
                          cosmetic changes
02-APR-1998  MGD  v2.4.0, report file locked against access a non-fatal error,
                          added form-based path to override path-info,
                          refined (some might say corrected ;^) HTML tag parsing
19-FEB-1998  MGD  v2.3.2, bugfix; "&hits=document" nesting,
                          modified "?about=search" redirection
19-AUG-1997  MGD  v2.3.1, MapUrl() to MapUrl_Map() for conditional mapping
23-MAY-1997  MGD  v2.3.0, wildcard search,
                          un-escape entities (e.g. "&lt;") before match
19-SEP-1995  MGD  v2.2.1, replace <CR><LF> carriage-control with single <LF>,
                          still acceptable for HTTP, slightly more efficient
24-MAY-1995  MGD  v2.2.0, minor changes for AXP compatibility
21-APR-1995  MGD  v2.1.1, added 'FormWhat'
27-MAR-1995  MGD  v2.1.0, modifications to CGI interface
05-DEC-1994  MGD  v2.0.0, major revision, URL mapping, CGI-like interface
10-JUN-1994  MGD  v1.0.0, initial development
*/
/*****************************************************************************/

#define SOFTWAREVN "4.2.0"
#define SOFTWARENM "QUERY"
#define SOFTWARECR "Copyright (C) 1996-2020 Mark G.Daniel"
#ifdef __ALPHA
#  define SOFTWAREID SOFTWARENM " AXP-" SOFTWAREVN
#endif
#ifdef __ia64
#  define SOFTWAREID SOFTWARENM " IA64-" SOFTWAREVN
#endif
#ifdef __VAX
#  define SOFTWAREID SOFTWARENM " VAX-" SOFTWAREVN
#endif
#ifdef __x86_64
#  define SOFTWAREID SOFTWARENM " X86-" SOFTWAREVN
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/* VMS-related header files */
#include <descrip.h>
#include <lib$routines.h>
#include <libdef.h>
#include <rmsdef.h>
#include <rms.h>
#include <ssdef.h>
#include <stsdef.h>
#include <syidef.h>
#include <starlet.h>

/* application header files */
#include "enamel.h"
#include "query.h"
#include "cgilib.h"

/* mainly to allow easy use of the __unaligned directive */
#define ULONGPTR __unaligned unsigned long*
#define USHORTPTR __unaligned unsigned short*
#define INT64PTR __unaligned __int64*

#ifndef __VAX
#   pragma nomember_alignment
#endif

#ifndef __VAX
#  ifndef NO_ODS_EXTENDED
#     define ODS_EXTENDED 1
      /* this is smaller than the technical maximum, but still quite large! */
#     define ODS_MAX_FILE_NAME_LENGTH 511
#     define ODS_MAX_FILESYS_NAME_LENGTH 264
#  endif
#endif
#define ODS2_MAX_FILE_NAME_LENGTH 255
#ifndef ODS_MAX_FILE_NAME_LENGTH
#  define ODS_MAX_FILE_NAME_LENGTH ODS2_MAX_FILE_NAME_LENGTH
#endif
#if ODS_MAX_FILE_NAME_LENGTH < ODS2_MAX_FILE_NAME_LENGTH
#  define ODS_MAX_FILE_NAME_LENGTH ODS2_MAX_FILE_NAME_LENGTH
#endif

#define BOOL int
#define true 1
#define false 0
 
#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) !(((x) & STS$M_SUCCESS))

#define FI_LI __FILE__, __LINE__

/* convenience macro, returns a pointer to calloc()ed HTML-escaped string */
#define HTMLESC(str) ((char*)CgiLibHtmlEscape(str,-1,NULL,-1))

char  DefaultHtmlTypes [] = DEFAULT_HTML_TYPES,
      DefaultStyle [] = DEFAULT_QUERY_STYLE,
      DefaultTextTypes [] = DEFAULT_TEXT_TYPES,
      Utility [] = "QUERY";

BOOL  CaseSensitive,
      Debug,
      DocumentOnly,
      ExactNumberOfRecords,
      ThereHasBeenOutput,
      ListStarted,
      TreatHtmlAsPlain,
      FormatLikeVms;

int  ExtractNumberOfRecords,
     KeyCount,
     FileCount,
     FileHitCount,
     ResFileNameLength,
     NotSearchedFileCount,
     OdsExtended,
     TotalHitCount,
     RecordCount,
     SearchStringLength;

char  DocumentName [2048],
      ExpFileName [ODS_MAX_FILE_NAME_LENGTH+1],
      FileNamePath [ODS_MAX_FILE_NAME_LENGTH+1],
      FormHtml [256],
      FormPath [256],
      FormText [256],
      KeyName [16],
      ResFileName [ODS_MAX_FILE_NAME_LENGTH+1],
      SearchString [256],
      SoftwareCopy [] = SOFTWARECR,
      SoftwareID [48],
      UrlEncodedFileNamePath [ODS_MAX_FILE_NAME_LENGTH+1],
      UrlEncodedSearchString [256];

char  *AboutPathPtr = DEFAULT_ABOUT_PATH,
      *CgiEnvironmentPtr,
      *CgiFormAboutPtr,
      *CgiFormCasePtr,
      *CgiFormExactPtr,
      *CgiFormExtractPtr,
      *CgiFormHitsPtr,
      *CgiFormHtmlPtr,
      *CgiFormPathPtr,
      *CgiFormPlainPtr,
      *CgiFormSearchPtr,
      *CgiFormTextPtr,
      *CgiFormTargetPtr,
      *CgiFormWhatPtr,
      *CgiKeyCountPtr,
      *CgiKeyValuePtr,
      *CgiPathInfoPtr,
      *CgiPathTranslatedPtr,
      *CgiQueryStringPtr,
      *CgiRequestMethodPtr,
      *CgiRequestUriPtr,
      *CgiScriptNamePtr,
      *CgiServerSoftwarePtr,
      *CharsetPtr,
      *CliCharsetPtr,
      *ExtractScriptNamePtr,
      *FormPlainPtr,
      *HtmlFileTypesPtr = DefaultHtmlTypes,
      *LastModifiedTimeFormatPtr = DEFAULT_STRFTIME_FORMAT,
      *RequeryPtr,
      *StyleSheetPtr = "",
      *TargetPtr = "",
      *TextFileTypesPtr = DefaultTextTypes;
      
   struct FAB  SearchFab;
   struct NAM  SearchNam;
#ifdef ODS_EXTENDED
   struct NAML  SearchNaml;
#endif /* ODS_EXTENDED */

/* required prototypes */
int QueryHtmlEscape (char*, int, char*, int);
char* SearchTextString (char*, char*, BOOL, BOOL, int*);

/*****************************************************************************/
/*
'argc/argv' are only required to support OSU, in particular CgiLibOsuInit().
*/

main
(
int argc,
char *argv[]
)
{
   int  status,
        Count;
   char  *cptr;

   /*********/
   /* begin */
   /*********/

   sprintf (SoftwareID, "%s (%s)", SOFTWAREID, CgiLibEnvironmentVersion());

   if (getenv ("QUERY$DBUG") != NULL) Debug = true;
   if (Debug)
   {
      CgiLibResponseHeader (200, "text/plain");
      CgiLibEnvironmentSetDebug (Debug);
   }

   CgiLibEnvironmentInit (argc, argv, false);
   CgiEnvironmentPtr = CgiLibEnvironmentName ();

   GetParameters ();

   CgiLibResponseSetCharset (CliCharsetPtr);
   CgiLibResponseSetSoftwareID (SoftwareID);
   CgiLibResponseSetErrorMessage ("Reported by WASDquery");

   if (StyleSheetPtr[0])
   {
      cptr = calloc (1, 64+strlen(StyleSheetPtr));
      sprintf (cptr, "<link rel=\"stylesheet\" \
type=\"text/css\" href=\"%s\">\n",
               StyleSheetPtr);
      StyleSheetPtr = cptr;
   }

#ifdef ODS_EXTENDED
   OdsExtended = (GetVmsVersion() >= 72);
   ENAMEL_NAML_SANITY_CHECK
#endif /* ODS_EXTENDED */

   CgiServerSoftwarePtr = CgiLibVar ("WWW_SERVER_SOFTWARE");
   CgiRequestMethodPtr = CgiLibVar ("WWW_REQUEST_METHOD");
   if (strcmp (CgiRequestMethodPtr, "GET"))
   {
      CgiLibResponseHeader (501, "text/html");
      fprintf (stdout, "Not implemented!\n");
      return;
   }

   CgiScriptNamePtr = CgiLibVar ("WWW_SCRIPT_NAME");
   CgiPathInfoPtr = CgiLibVar ("WWW_PATH_INFO");
   CgiQueryStringPtr = CgiLibVar ("WWW_QUERY_STRING");
   if (!CgiQueryStringPtr[0])
   {
      QueryForm ();
      return;
   }

   /* script name and any path */
   CgiRequestUriPtr = CgiLibVar ("WWW_REQUEST_URI");
   for (cptr = CgiRequestUriPtr; *cptr && *cptr != '?'; cptr++);
   RequeryPtr = calloc (1, cptr - CgiRequestUriPtr + 1);
   memcpy (RequeryPtr, CgiRequestUriPtr, cptr - CgiRequestUriPtr);

   /* kludge around a path not being MAP=ELLIPSIS */
   cptr = CgiPathTranslatedPtr = CgiLibVar ("WWW_PATH_TRANSLATED");
   while (*cptr)
   {
      if (cptr[0] == '|' && cptr[1] == '|' && cptr[2] == '|' && cptr[3] != '|')
      {
         cptr[0] = cptr[1] = cptr[2] = '.';
         break;
      }
      cptr++;
   }

   CgiFormAboutPtr = CgiLibVar ("WWW_FORM_ABOUT");
   CgiFormCasePtr = CgiLibVar ("WWW_FORM_CASE");
   CgiFormExactPtr = CgiLibVar ("WWW_FORM_EXACT");
   CgiFormExtractPtr = CgiLibVar ("WWW_FORM_EXTRACT");
   CgiFormHitsPtr = CgiLibVar ("WWW_FORM_HITS");
   CgiFormHtmlPtr = CgiLibVar ("WWW_FORM_HTML");
   CgiFormPathPtr = CgiLibVar ("WWW_FORM_PATH");
   CgiFormPlainPtr = CgiLibVar ("WWW_FORM_PLAIN");
   CgiFormSearchPtr = CgiLibVar ("WWW_FORM_SEARCH");
   CgiFormTargetPtr = CgiLibVar ("WWW_FORM_TARGET");
   CgiFormTextPtr = CgiLibVar ("WWW_FORM_TEXT");
   CgiFormWhatPtr = CgiLibVar ("WWW_FORM_WHAT");
   CgiKeyCountPtr = CgiLibVar ("WWW_KEY_COUNT");

   SearchString[0] = '\0';
   KeyCount = atoi(CgiKeyCountPtr);
   for (Count = 1; Count <= KeyCount; Count++)
   {
      sprintf (KeyName, "WWW_KEY_%d", Count);
      CgiKeyValuePtr = CgiLibVar (KeyName);
      if (SearchString[0]) strcat (SearchString, " ");
      strcat (SearchString, CgiKeyValuePtr);
   }
   if (SearchString[0])
      CgiFormSearchPtr = SearchString;
   else
      strcpy (SearchString, CgiFormSearchPtr);

   if (strsame (CgiFormAboutPtr, "SEARCH", 6))
   {
      CgiLibResponseRedirect (AboutPathPtr);
      exit (SS$_NORMAL);
   }

   if (CgiFormPathPtr[0])
   {
      /* path supplied indirectly via form field, redirect to translate */
      CgiScriptNamePtr = CgiLibVar ("WWW_SCRIPT_NAME");
      CgiQueryStringPtr = CgiLibVar ("WWW_QUERY_STRING");

      CgiLibResponseRedirect ("%s%s%s%s",
         CgiScriptNamePtr, CgiFormPathPtr,
         CgiQueryStringPtr[0] ? "?" : "", CgiQueryStringPtr);

      exit (SS$_NORMAL);
   }

   if (!CgiPathTranslatedPtr[0])
   {
      CgiLibResponseError (FI_LI, 0, "Search path not supplied.");
      exit (SS$_NORMAL);
   }
   if (!CgiFormSearchPtr[0])
   {
      CgiLibResponseError (FI_LI, 0, "Search string not supplied.");
      exit (SS$_NORMAL);
   }
   if (CgiFormExtractPtr[0] && !isdigit(CgiFormExtractPtr[0]))
   {
      CgiLibResponseError (FI_LI, 0, "Invalid number of lines to extract.");
      exit (SS$_NORMAL);
   }

   ExtractNumberOfRecords = atoi(CgiFormExtractPtr);
   if (!ExtractNumberOfRecords)
      ExtractNumberOfRecords = DEFAULT_EXTRACT_RECORDS;

   if (toupper(CgiFormExactPtr[0]) == 'Y')
      ExactNumberOfRecords = true;
   else
      ExactNumberOfRecords = false;

   if (toupper(CgiFormCasePtr[0]) == 'Y')
      CaseSensitive = true;
   else
      CaseSensitive = false;

   if (toupper(CgiFormHitsPtr[0]) == 'D')
      DocumentOnly = true;
   else
      DocumentOnly = false;

   if (toupper(CgiFormPlainPtr[0]) == 'Y')
   {
      TreatHtmlAsPlain = true;
      FormPlainPtr = "&plain=yes";
   }
   else
   {
      TreatHtmlAsPlain = false;
      FormPlainPtr = "";
   }

   if (CgiFormHtmlPtr[0])
   {
      HtmlFileTypesPtr = CgiFormHtmlPtr;
      sprintf (FormHtml, "&html=%s", CgiFormHtmlPtr);
   }
   if (CgiFormTextPtr[0])
   {
      TextFileTypesPtr = CgiFormTextPtr;
      sprintf (FormText, "&text=%s", CgiFormTextPtr);
   }

   if (ExtractScriptNamePtr == NULL)
   {
      if (CgiLibEnvironmentIsWasd() ||
          CgiLibEnvironmentIsCgiPlus())
         ExtractScriptNamePtr = DEFAULT_WASD_EXTRACT;
      else
      if (CgiLibEnvironmentIsOsu())
         ExtractScriptNamePtr = DEFAULT_OSU_EXTRACT;
      else
         ExtractScriptNamePtr = DEFAULT_CGI_EXTRACT;
   }

   if (CgiFormTargetPtr[0])
   {
      TargetPtr = calloc (1, 32+strlen(CgiFormTargetPtr));
      sprintf (TargetPtr, " target=\"%s\"", CgiFormTargetPtr);
   }

   SearchFiles ();

   exit (SS$_NORMAL);
}

/*****************************************************************************/
/*
Get "command-line" parameters, whether from the command-line or from a
configuration symbol or logical containing the equivalent.  OSU scripts have
the 'method', 'url' and 'protocol' supplied as P1, P2, P3 (these being detected
and used by CGILIB), and are of no interest to this function.
*/

GetParameters ()

{
   static char  CommandLine [256];
   static unsigned long  Flags = 0;

   int  status,
        SkipParameters;
   unsigned short  Length;
   char  ch;
   char  *aptr, *cptr, *clptr, *sptr;
   $DESCRIPTOR (CommandLineDsc, CommandLine);

   /*********/
   /* begin */
   /*********/

   if ((clptr = getenv ("QUERY$PARAM")) == NULL)
   {
      /* get the entire command line following the verb */
      if (VMSnok (status =
          lib$get_foreign (&CommandLineDsc, 0, &Length, &Flags)))
         exit (status);
      (clptr = CommandLine)[Length] = '\0';
   }

   /* if [C]SWS (VMS Apache) */
   if (CgiLibEnvironmentIsApache())
   {
      /* CSWS 1.2/3 look for something non-space outside of quotes */
      for (cptr = clptr; *cptr; cptr++)
      {
         if (isspace(*cptr)) continue;
         if (*cptr != '\"') break;
         cptr++;
         while (*cptr && *cptr != '\"') cptr++;
         if (*cptr) cptr++;
      }
      /* CSWS 1.2/3 if nothing outside of quotes then ignore command line */
      if (!*cptr) return;
      /* SWS 2.0 doesn't begin with /APACHE from DCL procedure wrapper */
      if (!strsame (cptr, "/APACHE", 7)) return;
   }

   /* if OSU environment then skip P1, P2, P3 */
   if (CgiLibEnvironmentIsOsu())
      SkipParameters = 3;
   else
      SkipParameters = 0;

   aptr = NULL;
   ch = *clptr;
   for (;;)
   {
      if (aptr != NULL) *aptr = '\0';
      if (!ch) break;

      *clptr = ch;
      if (Debug) fprintf (stdout, "clptr |%s|\n", clptr);
      while (*clptr && isspace(*clptr)) *clptr++ = '\0';
      aptr = clptr;
      if (*clptr == '/') clptr++;
      while (*clptr && !isspace (*clptr) && *clptr != '/')
      {
         if (*clptr != '\"')
         {
            clptr++;
            continue;
         }
         cptr = clptr;
         clptr++;
         while (*clptr)
         {
            if (*clptr == '\"')
               if (*(clptr+1) == '\"')
                  clptr++;
               else
                  break;
            *cptr++ = *clptr++;
         }
         *cptr = '\0';
         if (*clptr) clptr++;
      }
      ch = *clptr;
      if (*clptr) *clptr = '\0';
      if (Debug) fprintf (stdout, "aptr |%s|\n", aptr);
      if (!*aptr) continue;

      if (SkipParameters)
      {
         SkipParameters--;
         continue;
      }

      if (strsame (aptr, "/APACHE", 4))
      {
         /* just skip this marker for command-line parameters under SWS 2.0 */
         continue;
      }
      if (strsame (aptr, "/ABOUT=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         AboutPathPtr = cptr;
         continue;
      }
      if (strsame (aptr, "/ADDHTML=", 7))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         sptr = malloc (strlen(HtmlFileTypesPtr)+strlen(cptr)+2);
         strcpy (sptr, HtmlFileTypesPtr);
         strcat (sptr, ",");
         strcat (sptr, cptr);
         HtmlFileTypesPtr = sptr;
         continue;
      }
      if (strsame (aptr, "/ADDTEXT=", 7))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         sptr = malloc (strlen(TextFileTypesPtr)+strlen(cptr)+2);
         strcpy (sptr, TextFileTypesPtr);
         strcat (sptr, ",");
         strcat (sptr, cptr);
         TextFileTypesPtr = sptr;
         continue;
      }
      if (strsame (aptr, "/CHARSET=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         CliCharsetPtr = cptr;
         continue;
      }
      if (strsame (aptr, "/DBUG", -1))
      {
         Debug = true;
         continue;
      }
      if (strsame (aptr, "/EXTRACT=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         ExtractScriptNamePtr = cptr;
         continue;
      } 
      if (strsame (aptr, "/HTML=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         HtmlFileTypesPtr = cptr;
         continue;
      } 
      if (strsame (aptr, "/ODS5", 5))
      {
         OdsExtended = true;
         continue;
      }
      if (strsame (aptr, "/NOODS5", 7))
      {
         OdsExtended = false;
         continue;
      }
      if (strsame (aptr, "/STYLE=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         StyleSheetPtr = cptr;
         continue;
      }
      if (strsame (aptr, "/TEXT=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         TextFileTypesPtr = cptr;
         continue;
      }
      if (strsame (aptr, "/TIMEFMT=", 4))
      {
         for (cptr = aptr; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         LastModifiedTimeFormatPtr = cptr;
         continue;
      } 

      if (*aptr != '/')
      {
         fprintf (stdout, "%%%s-E-MAXPARM, too many parameters\n \\%s\\\n",
                  Utility, aptr);
         exit (STS$K_ERROR | STS$M_INHIB_MSG);
      }
      else
      {
         fprintf (stdout, "%%%s-E-IVQUAL, unrecognized qualifier\n \\%s\\\n",
                  Utility, aptr+1);
         exit (STS$K_ERROR | STS$M_INHIB_MSG);
      }
   }
}

/*****************************************************************************/
/*
Search files in specification for specified string.  Unless a file type 
(extension) matches an HTML type then consider the file to be a plain-text 
file.  The two need to searched differently because of the HTML markup tags 
compsrising an HTML file.
*/ 

SearchFiles ()

{
   static int  TimerElapsed = 1,
               TimerCpu = 2,
               TimerBio = 3,
               TimerDio = 4;
   static $DESCRIPTOR (ElapsedTimeFaoDsc, "!%T");
   static $DESCRIPTOR (StatisticsFaoDsc,
"Elapsed: !AZ &nbsp;CPU: !2ZL:!2ZL.!2ZL &nbsp;I/O: !UL &nbsp;Disk: !UL \
&nbsp;Records (lines): !UL");

   int  status,
        BioCount,
        CpuTime,
        DioCount,
        ExpandedNameOffset,
        ResultVersionLength;
   unsigned long  ResultFnb;
   unsigned long  ElapsedTime [2];
   unsigned short  Length;
   char  *cptr, *sptr,
         *FilePtr,
         *FileHitPtr,
         *FileNamePathPtr,
         *ResultNamePtr,
         *ResultTypePtr,
         *ResultVersionPtr,
         *TotalHitPtr;
   char  ElapsedString [32],
         EscapedPath [ODS_MAX_FILE_NAME_LENGTH+1],
         EscapedSearchString [256],
         NotSearchedString [32],
         Statistics [256],
         String [1024];
   $DESCRIPTOR (ElapsedStringDsc, ElapsedString);
   $DESCRIPTOR (StatisticsDsc, Statistics);

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SearchFiles() |%s|%s|\n",
               CgiPathTranslatedPtr, SearchString);

   lib$init_timer (0);

   ListStarted = false;

   FormatLikeVms = false;
   for (cptr = CgiPathInfoPtr; *cptr; cptr++);
   while (cptr > CgiPathInfoPtr && *cptr != '/')
   {
      if (*cptr == ';')
      {
         FormatLikeVms = true;
         break;
      }
      cptr--;
   }

   if (FormatLikeVms)
      CgiLibHtmlEscape (CgiPathTranslatedPtr, -1,
                        EscapedPath, sizeof(EscapedPath));
   else
   {
      if (!CgiPathInfoPtr[0]) CgiPathInfoPtr = "/";
      CgiLibHtmlEscape (CgiPathInfoPtr, -1, EscapedPath, sizeof(EscapedPath));
   }

   if (SearchString[0])
   {
      SearchStringLength = strlen(SearchString);
      CgiLibUrlEncode (SearchString, -1, UrlEncodedSearchString, -1);
      CgiLibHtmlEscape (SearchString, -1,
                        EscapedSearchString, sizeof(EscapedSearchString));
   }
   else
   {
      SearchStringLength = 0;
      EscapedSearchString[0] = UrlEncodedSearchString[0] = '\0';
   }

   sptr = String;
   if (CgiFormWhatPtr[0])
      sptr += sprintf (sptr, "Search %s for &quot;%s&quot;",
                       CgiFormWhatPtr, EscapedSearchString);
   else
      sptr += sprintf (sptr, "Search %s for &quot;%s&quot;",
                        EscapedPath, EscapedSearchString);

   SearchFab = cc$rms_fab;
   SearchFab.fab$l_dna = "*.*;0";
   SearchFab.fab$b_dns = 5;

#ifdef ODS_EXTENDED
   if (OdsExtended)
   {
      SearchFab.fab$l_fna = (char*)-1;
      SearchFab.fab$b_fns = 0;
      SearchFab.fab$l_nam = (struct namdef*)&SearchNaml;

      ENAMEL_RMS_NAML(SearchNaml)
      SearchNaml.naml$l_long_filename = CgiPathTranslatedPtr;
      SearchNaml.naml$l_long_filename_size = strlen(CgiPathTranslatedPtr);
      SearchNaml.naml$l_long_expand = ExpFileName;
      SearchNaml.naml$l_long_expand_alloc = sizeof(ExpFileName)-1;
      SearchNaml.naml$l_long_result = ResFileName;
      SearchNaml.naml$l_long_result_alloc = sizeof(ResFileName)-1;
   }
   else
#endif /* ODS_EXTENDED */
   {
      SearchFab.fab$l_fna = CgiPathTranslatedPtr;
      SearchFab.fab$b_fns = strlen(CgiPathTranslatedPtr);
      SearchFab.fab$l_nam = &SearchNam;

      SearchNam = cc$rms_nam;
      SearchNam.nam$l_esa = ExpFileName;
      SearchNam.nam$b_ess = ODS2_MAX_FILE_NAME_LENGTH;
      SearchNam.nam$l_rsa = ResFileName;
      SearchNam.nam$b_rss = ODS2_MAX_FILE_NAME_LENGTH;
   }

   if (VMSnok (status = sys$parse (&SearchFab, 0, 0)))
   {
      CgiLibResponseError (FI_LI, status, HTMLESC(CgiPathInfoPtr));
      return (status);
   }

   if (Debug) fprintf (stdout, "ExpFileName |%s|\n", ExpFileName);

   /* establish the base for generated file name used in the file path */
#ifdef ODS_EXTENDED
   if (OdsExtended)
      sptr = SearchNaml.naml$l_long_name - 1;
   else
#endif /* ODS_EXTENDED */
      sptr = SearchNam.nam$l_name - 1;
   cptr = ExpFileName;
   while (cptr < sptr)
   {
      if (*cptr == '*' || *cptr == '%' ||
          (cptr[0] == '.' && cptr[1] == '.' && cptr[2] == '.'))
      {
         if (cptr[0] != '.' || cptr[1] != '.' || cptr[2] != '.') cptr--;
         break;
      }
      cptr++;
   }
   while (cptr > ExpFileName && *cptr != '[' && *cptr != '.' && *cptr != ']')
      cptr--;
   ExpandedNameOffset = cptr - ExpFileName + 1;

   /* establish the base for generated file path */
   sptr = FileNamePath;
   cptr = CgiPathInfoPtr;
   while (*cptr)
   {
      if (*cptr == '*' || *cptr == '%')
      {
         cptr--;
         sptr--;
         break;
      }
      if (cptr[0] == '.' && cptr[1] == '.' && cptr[2] == '.') break;
      *sptr++ = *cptr++;
   }
   if ((cptr[0] == '.' && cptr[1] == '.' && cptr[2] == '.') &&
       sptr[-1] != '/' && sptr[-1] != '*' && sptr[-1] != '%')
      *sptr++ = '/';
   else
   {
      while (sptr > FileNamePath && *sptr != '/') sptr--;
      *sptr++;
   }
   *sptr = '\0';
   FileNamePathPtr = sptr;
   if (Debug) fprintf (stdout, "FileNamePath |%s|\n", FileNamePath);

   /*************************/
   /* begin the page output */
   /*************************/

   CgiLibResponseHeader (200, "text/html");

   fprintf (stdout,
"<!DOCTYPE html>\n\
<html>\n\
<head>\n\
<meta http-equiv=\"X-UA-Compatible\" content=\"IE=edge\">\n\
<meta name=\"generator\" content=\"%s\">\n\
<meta name=\"copyright\" content=\"%s - GPL licensed\">\n\
<meta name=\"environment\" content=\"%s\">\n\
<meta name=\"text-types\" content=\"%s\">\n\
<meta name=\"html-types\" content=\"%s\">\n\
<title>%s</title>\n\
<style type=\"text/css\">\n\
%s\
</style>\n\
%s\
</head>\n\
<body>\n",
      SoftwareID, SoftwareCopy,
      CgiEnvironmentPtr,
      TextFileTypesPtr,
      HtmlFileTypesPtr,
      String,
      DefaultStyle,
      StyleSheetPtr);

   fprintf (stdout, "<div class=\"header\">%s</div>\n", String);
 
   fflush (stdout);

   ThereHasBeenOutput = true;

   /***************/
   /* file search */
   /***************/

   for (;;)
   {
      status = sys$search (&SearchFab, 0, 0);
      if (Debug) fprintf (stdout, "sys$search() %%X%08.08X\n", status);
      if (status == RMS$_PRV) continue;
      if (VMSnok (status)) break;

#ifdef ODS_EXTENDED
      if (OdsExtended)
      {
         ResultFnb = SearchNaml.naml$l_fnb;
         ResultNamePtr = SearchNaml.naml$l_long_name;
         ResultTypePtr = SearchNaml.naml$l_long_type;
         ResultVersionPtr = SearchNaml.naml$l_long_ver;
         ResultVersionLength = SearchNaml.naml$l_long_ver_size;
         ResFileNameLength = SearchNaml.naml$l_long_ver -
                             SearchNaml.naml$l_long_result;
      }
      else
#endif /* ODS_EXTENDED */
      {
         ResultFnb = SearchNam.nam$l_fnb;
         ResultNamePtr = SearchNam.nam$l_name;
         ResultTypePtr = SearchNam.nam$l_type;
         ResultVersionPtr = SearchNam.nam$l_ver;
         ResultVersionLength = SearchNam.nam$b_ver;
         ResFileNameLength = SearchNam.nam$l_ver -
                             SearchNam.nam$l_rsa;
      }

      if (FormatLikeVms)
      {
         ResultVersionPtr[ResultVersionLength] = '\0';
         ResFileNameLength += ResultVersionLength;
      }
      else
         ResultVersionPtr[0] = '\0';
      if (Debug)
         fprintf (stdout, "ResFileName %d |%s|\n",
                  ResFileNameLength, ResFileName);

      /* generate a file path from the base file path and base file name */
      sptr = FileNamePathPtr;
      cptr = ResFileName + ExpandedNameOffset;
      while (*cptr && *cptr != ']' && cptr < ResultNamePtr)
      {
         if (CgiLibEnvironmentIsWasd() &&
             !memcmp (cptr, "000000.", 7) ||
             !memcmp (cptr, "000000]", 7)) cptr += 7;
         if (*cptr == '.')
         {
            *sptr++ = '/';
            cptr++;
         }
         else
            *sptr++ = tolower(*cptr++);
      }
      if (*cptr == ']')
      {
         cptr++;
         *sptr++ = '/';
      }
      while (*cptr) *sptr++ = tolower(*cptr++);
      *sptr = '\0';
      if (Debug) fprintf (stdout, "FileNamePath |%s|\n", FileNamePath);

      if (SameFileType (HtmlFileTypesPtr, ResultTypePtr+1))
      {
         /* if HTML document's <title> can't be resolved then use file name */
#ifdef ODS_EXTENDED
         if (OdsExtended)
            cptr = (char*)CgiLibUrlEncodeFileName (ResultNamePtr, NULL, 0,
                                                   FormatLikeVms,
                                                   !FormatLikeVms);
         else
#endif /* ODS_EXTENDED */
            cptr = ResultNamePtr;

         sprintf (DocumentName,
"%s <span class=\"filename\">[file name]</span>",
                  cptr);

         if (TreatHtmlAsPlain)
         {
            if (VMSok (status = SearchTextFile ()))
               FileCount++;
            else
            if (status != RMS$_PRV &&
                status != RMS$_WLK)
            {
               CgiLibResponseError (FI_LI, status, FileNamePath);
               return (SS$_NORMAL);
            }
         }
         else
         {
            if (VMSok (status = SearchHtmlFile ()))
               FileCount++;
            else
            if (status != RMS$_PRV &&
                status != RMS$_WLK)
            {
               CgiLibResponseError (FI_LI, status, FileNamePath);
               return (SS$_NORMAL);
            }
         }
      }
      else
      if (SameFileType (TextFileTypesPtr, ResultTypePtr+1))
      {
#ifdef ODS_EXTENDED
         if (OdsExtended)
            cptr = (char*)CgiLibUrlEncodeFileName (ResultNamePtr, NULL, 0,
                                                   FormatLikeVms,
                                                   !FormatLikeVms);
         else
#endif /* ODS_EXTENDED */
            cptr = ResultNamePtr;

         strcpy (DocumentName, cptr);

         if (VMSok (status = SearchTextFile ()))
            FileCount++;
         else
         if (status != RMS$_PRV &&
             status != RMS$_WLK)
         {
            CgiLibResponseError (FI_LI, status, FileNamePath);
            return (SS$_NORMAL);
         }
      }
      else
         NotSearchedFileCount++;

      if (!FormatLikeVms)
      {
         /* restore the version delimiter */
         ResultVersionPtr[0] = ';';
      }
   }

   if (Debug) fprintf (stdout, "sys$search() %%X%08.08X\n", status);

   /*******************/
   /* end file search */
   /*******************/

   /* if its a search list treat directory not found as if file not found */
   if ((ResultFnb & NAM$M_SEARCH_LIST) && status == RMS$_DNF)
      status = RMS$_FNF;
   if (status == RMS$_FNF || status == RMS$_NMF) status = SS$_NORMAL;
   if (VMSnok (status))
   {
       CgiLibResponseError (FI_LI, status, HTMLESC(CgiPathInfoPtr)); 
       return (SS$_NORMAL);
   }

   /*********************/
   /* results of search */
   /*********************/

   lib$stat_timer (&TimerElapsed, &ElapsedTime, 0);
   lib$stat_timer (&TimerCpu, &CpuTime, 0);
   lib$stat_timer (&TimerBio, &BioCount, 0);
   lib$stat_timer (&TimerDio, &DioCount, 0);

   sys$fao (&ElapsedTimeFaoDsc, &Length, &ElapsedStringDsc, &ElapsedTime);
   ElapsedString[Length] = '\0';
   sys$fao (&StatisticsFaoDsc, &Length, &StatisticsDsc,
            ElapsedString+3, CpuTime/6000, CpuTime/100, CpuTime%100,
            BioCount, DioCount, RecordCount);
   Statistics[Length] = '\0';

   if (!NotSearchedFileCount)
      NotSearchedString[0] = '\0';
   else
   if (NotSearchedFileCount == 1)
      strcpy (NotSearchedString, " (1 not)");
   else
      sprintf (NotSearchedString, " (%d not)", NotSearchedFileCount);

   if (ListStarted) fputs ("</ol>\n", stdout);

   fputs ("<div class=\"footer\">\n\
<div class=\"summary\">", stdout);

   if (TotalHitCount)
   {
      if (FileCount == 1) FilePtr = "file"; else FilePtr = "files";
      if (FileHitCount == 1) FileHitPtr = "file"; else FileHitPtr = "files";
      if (TotalHitCount == 1) TotalHitPtr = "hit"; else TotalHitPtr = "hits";
      if (DocumentOnly)
      {
         fprintf (stdout, "%d %s searched%s with %d %s hit.",
                  FileCount, FilePtr, NotSearchedString, FileHitCount,
                  FileHitPtr);
      }
      else
      {
         fprintf (stdout,
"%d %s searched%s with %d %s hit, for a total of %d %s.",
                  FileCount, FilePtr, NotSearchedString, FileHitCount,
                  FileHitPtr, TotalHitCount, TotalHitPtr);
      }
   }
   else
   {
      if (FileCount)
      {
         if (FileCount == 1) FilePtr = "file"; else FilePtr = "files";
         fprintf (stdout, "%d %s searched%s, string not found.",
                  FileCount, FilePtr, NotSearchedString);
      }
      else
         fprintf (stdout, "No files found!");
   }

   fprintf (stdout,
"</div>\n\
<div class=\"otherbutton\">\n\
<a href=\"%s\">Requery</a>\n\
<a target=\"_blank\" href=\"%s\">About</a>\n\
</div>\n\
<div class=\"statistics\">%s</div>\n\
</div>\n",
            RequeryPtr, AboutPathPtr, Statistics);

   if (VMSnok (status))
   {
       CgiLibResponseError (FI_LI, status, HTMLESC(CgiPathInfoPtr));
       return (SS$_NORMAL);
   }

   fprintf (stdout, "</body>\n</html>\n");

   return (status);
}

/*****************************************************************************/
/*
This function accepts a comma-separated list of (possibly wildcarded) file
types (extensions, e.g. "TXT,TEXT,COM,C,PAS,FOR,RPT*") and a VMS file type
(e.g. ".TXT;", ".TXT", "TXT").  Returns true if the file type is in the list,
false if not.
*/

BOOL SameFileType
(
char *TypeList,
char *TypePtr
)
{
   char  ch;
   char  *cptr, *sptr, *zptr;
   char  FileType [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "SameFileType() |%s|%s|\n", FileType, TypeList);

   zptr = (sptr = FileType) + sizeof(FileType)-1;
   for (cptr = TypePtr;
        *cptr && *cptr != ';' && sptr < zptr;
        *sptr++ = *cptr++);
   *sptr = '\0';
   cptr = TypeList;
   while (*cptr)
   {
      for (sptr = cptr; *sptr && *sptr != ','; sptr++);
      ch = *sptr;
      *sptr = '\0';
      if (Debug) fprintf (stdout, "|%s|%s|\n", FileType, cptr);
      if ((SearchTextString (FileType, cptr, false, false, NULL)) != NULL)
      {
         *sptr = ch;
         return (true);
      }
      if (*sptr = ch) sptr++;
      cptr = sptr;
   }
   return (false);
}

/*****************************************************************************/
/*
String search allowing wildcard "*" (matching any multiple characters) and "%" 
(matching any single character).  Returns NULL if not found or a pointer to
start of matched string.  Setting 'ImpliedWildcards' means the 'SearchFor'
string is processed as if enclosed by '*' wildcard characters.
*/ 

char* SearchTextString
( 
char *SearchIn,
char *SearchFor,
BOOL CaseSensitive,
BOOL ImpliedWildcards,
int *MatchedLengthPtr
)
{
   char  *cptr, *sptr, *inptr,
         *RestartCptr,
         *RestartInptr,
         *MatchPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SearchTextString() |%s|%s|\n", SearchIn, SearchFor);

   if (MatchedLengthPtr != NULL) *MatchedLengthPtr = 0;
   if (!*(cptr = SearchFor)) return (NULL);
   inptr = MatchPtr = SearchIn;

   if (ImpliedWildcards)
   {
      /* skip leading text up to first matching character (if any!) */
      if (*cptr != '*' && *cptr != '%')
      {
         if (CaseSensitive)
            while (*inptr && *inptr != *cptr) inptr++;
         else
            while (*inptr && toupper(*inptr) != toupper(*cptr)) inptr++;
         if (Debug && !*inptr) fprintf (stdout, "1. NOT matched!\n");
         if (!*inptr) return (NULL);
         cptr++;
         MatchPtr = inptr++;
      }
   }

   for (;;)
   {
      if (CaseSensitive)
      {
         while (*cptr && *inptr && *cptr == *inptr)
         {
            cptr++;
            inptr++;
         }
      }
      else
      {
         while (*cptr && *inptr && toupper(*cptr) == toupper(*inptr))
         {
            cptr++;
            inptr++;
         }
      }

      if (ImpliedWildcards)
      {
         if (!*cptr)
         {
            if (Debug) fprintf (stdout, "1. matched!\n");
            if (MatchedLengthPtr != NULL) *MatchedLengthPtr = inptr - MatchPtr;
            return (MatchPtr);
         }
      }
      else
      {
         if (!*cptr && !*inptr)
         {
            if (Debug) fprintf (stdout, "2. matched!\n");
            if (MatchedLengthPtr != NULL) *MatchedLengthPtr = inptr - MatchPtr;
            return (MatchPtr);
         }
         if (*cptr != '*' && *cptr != '%')
         {
            if (Debug && !*inptr) fprintf (stdout, "3. NOT matched!\n");
            return (NULL);
         }
      }

      if (*cptr != '*' && *cptr != '%')
      {
         if (!*inptr)
         {
            if (Debug) fprintf (stdout, "4. NOT matched!\n");
            return (NULL);
         }
         cptr = SearchFor;
         MatchPtr = ++inptr;
         continue;
      }

      if (*cptr == '%')
      {
         /* single char wildcard processing */
         if (!*inptr) break;
         cptr++;
         inptr++;
         continue;
      }

      /* asterisk wildcard matching */
      while (*cptr == '*') cptr++;

      /* an asterisk wildcard at end matches all following */
      if (!*cptr)
      {
         if (Debug) fprintf (stdout, "5. matched!\n");
         while (*inptr) inptr++;
         if (MatchedLengthPtr != NULL) *MatchedLengthPtr = inptr - MatchPtr;
         return (MatchPtr);
      }

      /* note the current position in the string (first after the wildcard) */
      RestartCptr = cptr;
      for (;;)
      {
         /* find first char in SearchIn matching char after wildcard */
         if (CaseSensitive)
            while (*inptr && *cptr != *inptr) inptr++;
         else
            while (*inptr && toupper(*cptr) != toupper(*inptr)) inptr++;
         /* if did not find matching char in SearchIn being searched */
         if (Debug && !*inptr) fprintf (stdout, "6. NOT matched!\n");
         if (!*inptr) return (NULL);
         /* note the current position in SearchIn being searched */
         RestartInptr = inptr;
         /* try to match the remainder of the string and SearchIn */
         if (CaseSensitive)
         {
            while (*cptr && *inptr && *cptr == *inptr)
            {
               cptr++;
               inptr++;
            }
         }
         else
         {
            while (*cptr && *inptr && toupper(*cptr) == toupper(*inptr))
            {
               cptr++;
               inptr++;
            }
         }
         /* if reached the end of both string and SearchIn - match! */
         if (ImpliedWildcards)
         {
            if (!*cptr)
            {
               if (Debug) fprintf (stdout, "7. matched!\n");
               if (MatchedLengthPtr != NULL)
                  *MatchedLengthPtr = inptr - MatchPtr;
               return (MatchPtr);
            }
         }
         else
         {
            if (!*cptr && !*inptr)
            {
               if (Debug) fprintf (stdout, "8. matched!\n");
               if (MatchedLengthPtr != NULL)
                  *MatchedLengthPtr = inptr - MatchPtr;
               return (MatchPtr);
            }
         }
         /* break to the external loop if we encounter another wildcard */
         if (*cptr == '*' || *cptr == '%') break;
         /* lets have another go */
         cptr = RestartCptr;
         /* starting the character following the previous attempt */
         inptr = MatchPtr = RestartInptr + 1;
      }
   }
}

/*****************************************************************************/
/*
Search an HTML marked up file.  Simply count the number of '<' and '>' 
characters, which should be balanced, and when not inside an HTML markup tag 
search the text.  As HTML files cannot easily have text extracted from within 
them without the results being unpredictable simply return the document as 
having the search string hit.  The HTML <title> tag (if present) is used as 
the document name.
*/ 

SearchHtmlFile ()

{
   BOOL  InsideApplet = false,
         InsideComment = false,
         InsideHead = false,
         InsideScript = false,
         InsideServer = false,
         InsideStyle = false,
         InsideTitle = false;
   int  status,
        ByteCount,
        HitCount = 0,
        MatchedLength,
        RecordNumber = 0,
        TagCharCount = 0;
   unsigned long  UnixTime;
   char  ch;
   char  *cptr, *dptr, *rptr, *sptr, *tptr,
         *CaseSensitivePtr = "",
         *FileNamePathPtr;
   char  FileSize [32],
         LastModifiedTime [64],
         Record [MAX_RECORD_SIZE+1],
         String [MAX_RECORD_SIZE*2],
         Text [MAX_RECORD_SIZE*2];
   struct tm  *UnixTmPtr;
   struct FAB  FileFab;
#ifdef ODS_EXTENDED
   struct NAML  FileNaml;
#endif /* ODS_EXTENDED */
   struct RAB  FileRab;
   struct XABDAT  FileXabDat;
   struct XABFHC  FileXabFhc;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SearchHtmlFile() |%s|%s|\n", ResFileName, SearchString);

   FileFab = cc$rms_fab;
   FileFab.fab$b_fac = FAB$M_GET;
   FileFab.fab$b_shr = FAB$M_SHRGET | FAB$M_SHRPUT;
   FileFab.fab$l_xab = &FileXabDat;
   FileFab.fab$l_fop = FAB$M_NAM;
   FileXabDat = cc$rms_xabdat;
   FileXabDat.xab$l_nxt = &FileXabFhc;
   FileXabFhc = cc$rms_xabfhc;

#ifdef ODS_EXTENDED
   if (OdsExtended)
   {
      FileFab.fab$l_fna = (char*)-1;
      FileFab.fab$b_fns = 0;
      FileFab.fab$l_nam = (struct namdef*)&FileNaml;

      ENAMEL_RMS_NAML(FileNaml)
      FileNaml.naml$l_long_filename = ResFileName;
      FileNaml.naml$l_long_filename_size = ResFileNameLength;
   }
   else
#endif /* ODS_EXTENDED */
   {
      FileFab.fab$l_fna = ResFileName;  
      FileFab.fab$b_fns = ResFileNameLength;
   }

   status = sys$open (&FileFab, 0, 0);
   if (Debug) fprintf (stdout, "sys$open() %%X%08.08X\n", status);

   if (VMSnok (status))
   {
      if (status == RMS$_PRV) return (status);
      if (!ListStarted)
      {
         ListStarted = true;
         fputs ("<ol>\n", stdout);
      }
      fprintf (stdout,
"<li><span class=\"hit\">ERROR opening file (%%X%08.08X) \
%s <!-- %s --></span>\n",
         status,
         FileNamePath,
         ResFileName);
      return (SS$_NORMAL);
   }

   FileRab = cc$rms_rab;
   FileRab.rab$l_fab = &FileFab;
   /* 2 buffers and read ahead performance option */
   FileRab.rab$b_mbf = 2;
   FileRab.rab$l_rop = RAB$M_RAH;
   FileRab.rab$l_ubf = Record;
   FileRab.rab$w_usz = sizeof(Record)-1;

   if (VMSnok (status = sys$connect (&FileRab, 0, 0)))
   {
      if (Debug) fprintf (stdout, "sys$connect() %%X%08.08X\n", status);
      sys$close (&FileFab, 0, 0);
      return (status);
   }

   if (CaseSensitive) CaseSensitivePtr = "&case=yes";

   /**********************/
   /* search all records */
   /**********************/

if (Debug) fprintf (stdout, "+++++0 %d\n", TagCharCount);
   while (VMSok (status = sys$get (&FileRab, 0, 0)))
   {
      RecordNumber++;
      if (!FileRab.rab$w_rsz) continue;
      Record[FileRab.rab$w_rsz] = '\0';
      if (Debug)
         fprintf (stdout, "Record |%s|\n%d %d %d %d %d %d %d %d\n",
                  Record, TagCharCount, InsideApplet, InsideComment,
                  InsideHead, InsideScript, InsideServer, InsideStyle,
                  InsideTitle);
if (Debug) fprintf (stdout, "+++++1 %d\n", TagCharCount);

      /* terminate on any carriage control that may be in the record */
      for (rptr = Record; *rptr && *rptr != '\r' && *rptr != '\n'; rptr++);
      *rptr = '\0';
      /* continue if empty line */
      if (!Record[0]) continue;

      /**************************************/
      /* retrieve text not inside HTML tags */
      /**************************************/

      tptr = Text;
      rptr = Record;
      while (*rptr)
      {
         if (InsideComment)
         {
            if (rptr[0] == '-' && rptr[1] == '-' && rptr[2] == '>')
            {
               InsideComment = false;
               rptr += 3;
            }
            else
               rptr++;
            continue;
         }
         if (*rptr == '<' && *(ULONGPTR)rptr == '<!--')
         {
            InsideComment = true;
            rptr += 4;
            continue;
         }

         if (InsideScript)
         {
            /* pretty much anything goes except... */
            if (strsame (rptr, "</script>", 9))
            {
               rptr += 8;
               InsideScript = false;
            }
            else
               rptr++;
            continue;
         }

         if (InsideStyle)
         {
            /* pretty much anything goes except... */
            if (strsame (rptr, "</style>", 8))
            {
               rptr += 7;
               InsideStyle = false;
            }
            else
               rptr++;
            continue;
         }

         /* less-thans are forbidden inside tags! */
         if ((TagCharCount & 1) && rptr[0] == '<') break;

         /* consider adjacent white-space when determining what is a tag */
         if ((rptr[0] == '<' && rptr[1] && !isspace(rptr[1])) ||
             ((TagCharCount & 1) && rptr[0] == '>'))
         {
            TagCharCount++;
            rptr++;
            if (TagCharCount & 1)
            {
               /* checks to detect start and end of some tags */
               if (tolower(*rptr) == 'a' &&
                   (strsame (rptr, "applet>", 7) ||
                    strsame (rptr, "applet ", 7)))
               {
                  rptr += 6;
                  InsideApplet = true;
               }
               else
               if (strsame (rptr, "body>", 5) ||
                   strsame (rptr, "body ", 4))
               {
                  rptr += 4;
                  InsideHead = false;
               }
               else
               if (tolower(*rptr) == 'h' && strsame (rptr, "head>", 5))
               {
                  rptr += 4;
                  InsideHead = true;
               }
               else
               if (tolower(*rptr) == 's')
               {
                  if (strsame (rptr, "script>", 7) ||
                      strsame (rptr, "script ", 7))
                  {
                     rptr += 6;
                     InsideScript = true;
                  }
                  else
                  if (strsame (rptr, "server>", 7) ||
                      strsame (rptr, "server ", 7))
                  {
                     rptr += 6;
                     InsideServer = true;
                  }
                  if (strsame (rptr, "style>", 6) ||
                      strsame (rptr, "style ", 6))
                  {
                     rptr += 5;
                     InsideStyle = true;
                  }
               }
               else
               if (tolower(*rptr) == 't' && strsame (rptr, "title>", 6))
               { 
                  rptr += 5;
                  dptr = DocumentName;
                  InsideTitle = true;
               }
               else
               if (*rptr == '/')
               {
                  if (strsame (rptr, "/applet>", 8))
                  {
                     rptr += 7;
                     InsideApplet = false;
                  }
                  else
                  if (strsame (rptr, "/head>", 6))
                  {
                     rptr += 5;
                     InsideHead = false;
                  }
                  else
                  if (strsame (rptr, "/script>", 8))
                  {
                     rptr += 7;
                     InsideScript = false;
                  }
                  else
                  if (strsame (rptr, "/style>", 7))
                  {
                     rptr += 6;
                     InsideStyle = false;
                  }
                  else
                  if (strsame (rptr, "/server>", 8))
                  {
                     rptr += 7;
                     InsideServer = false;
                  }
                  else
                  if (strsame (rptr, "/title>", 7))
                  {
                     rptr += 6;
                     *dptr = '\0';
                     InsideTitle = false;
                  }
               }
            }
         }
         else
         {
            if (InsideTitle)
            {
               if (dptr < DocumentName+255)
                  *dptr++ = *rptr++;
               else
                  rptr++;
            }
            else
            if (InsideApplet ||
                InsideHead ||
                InsideScript ||
                InsideServer ||
                InsideStyle)
            {
               rptr++;
            }
            else
            if (TagCharCount & 1)
               rptr++;
            else
            {
               if (*rptr == '&')
               {
                  if (strsame (rptr, "&lt;", 4))
                  {
                     *tptr++ = '<';
                     rptr += 4;
                  }
                  else
                  if (strsame (rptr, "&gt;", 4))
                  {
                     *tptr++ = '>';
                     rptr += 4;
                  }
                  else
                  if (strsame (rptr, "&amp;", 5))
                  {
                     *tptr++ = '&';
                     rptr += 5;
                  }
                  else
                  if (strsame (rptr, "&quot;", 6))
                  {
                     *tptr++ = '\"';
                     rptr += 6;
                  }
                  else
                  if (strsame (rptr, "&nbsp;", 6))
                  {
                     *tptr++ = ' ';
                     rptr += 6;
                  }
                  else
                  if (*(rptr+1) == '#')
                  {
                     for (cptr = rptr+2; *cptr && isdigit(*cptr); cptr++);
                     if (*cptr == ';')
                     {
                        ch = atoi(rptr+2) & 0xff;
                        *tptr++ = ch;
                        rptr = cptr + 1;
                     }
                     else
                        *tptr++ = *rptr++;
                  }
                  else
                     *tptr++ = *rptr++;
               }
               else
                  *tptr++ = *rptr++;
            }
         }

      }  /* while (*rptr) */
if (Debug) fprintf (stdout, "+++++2 %d\n", TagCharCount);
 
      *tptr = '\0';
      if (!Text[0]) continue;

      tptr = SearchTextString (Text, SearchString,
                               CaseSensitive, true,
                               &MatchedLength);
      if (tptr != NULL)
      {
         /********/
         /* hit! */
         /********/

         if (Debug) fprintf (stdout, "Hit |%s|\n", tptr);

         TotalHitCount++;

#ifdef ODS_EXTENDED
         if (OdsExtended)
            FileNamePathPtr = 
               (char*)CgiLibUrlEncodeFileName (FileNamePath, NULL, 0,
                                               FormatLikeVms, !FormatLikeVms);
         else
#endif /* ODS_EXTENDED */
            FileNamePathPtr = FileNamePath;

         if (!HitCount++)
         {
            /*************************************/
            /* first hit, entire document anchor */
            /*************************************/

            UnixTime = decc$fix_time (&FileXabDat.xab$q_rdt);
            UnixTmPtr = localtime (&UnixTime);
            if (!strftime (LastModifiedTime, sizeof(LastModifiedTime),
                           LastModifiedTimeFormatPtr, UnixTmPtr))
               strcpy (LastModifiedTime, "**ERROR**");

            /* true for non-VAR record formats, almost true for those :^) */
            if (FileXabFhc.xab$l_ebk)
               ByteCount = ((FileXabFhc.xab$l_ebk - 1) * 512) +
                           FileXabFhc.xab$w_ffb;
            else
               ByteCount = FileXabFhc.xab$w_ffb;
            if (ByteCount < 1000)
               sprintf (FileSize, "%d bytes", ByteCount);
            else
            if (ByteCount % 1000 < 500)
               sprintf (FileSize, "%d kB", ByteCount/1000);
            else
               sprintf (FileSize, "%d kB", (ByteCount/1000)+1);

            if (!ListStarted)
            {
               ListStarted = true;
               fputs ("<ol>\n", stdout);
            }

            fprintf (stdout,
"<li><span class=\"hit\"><a%s href=\"%s\">%s</a> \
<span class=\"document\">HTML &nbsp;%s &nbsp;%s</span></span>\n",
               TargetPtr, FileNamePathPtr, DocumentName,
               LastModifiedTime, FileSize);

            if (DocumentOnly) break;

            fprintf (stdout, "<ol>\n");
         }

         /***********************/
         /* file extract anchor */
         /***********************/

         sptr = String;
         strcpy (sptr, "<li><span class=\"ishtml\">");
         sptr += 25;
         /* absorb leading white-space */
         for (cptr = Text; *cptr && (*cptr == ' ' || *cptr == '\t'); cptr++);
         /* copy the record up to the first character of the search string */
         sptr += QueryHtmlEscape (cptr, tptr-cptr, sptr, -1);

         /* add the HTML anchor */
         sptr += sprintf (sptr,
"<a%s href=\"%s%s?highlight=%s&anchor=yes%s%s%s#athitof%d\">",
                          TargetPtr, ExtractScriptNamePtr, FileNamePathPtr,
                          UrlEncodedSearchString, FormText, FormHtml,
                          FormPlainPtr, HitCount);

         /* matched string, highlighted within the anchor */
         sptr += QueryHtmlEscape (tptr, MatchedLength, sptr, -1);
         strcpy (sptr, "</a>");
         sptr += 4;

         /* rest of record after the matched search string */
         sptr += QueryHtmlEscape (tptr+MatchedLength, -1, sptr, -1);

         strcpy (sptr, "</span>\n");
         sptr += 8;

         if (Debug) fprintf (stdout, "String |%s|\n", String);
         fputs (String, stdout);
      }
   }
if (Debug) fprintf (stdout, "+++++3 %d\n", TagCharCount);

   /***************/
   /* end of file */
   /***************/

   if (status == RMS$_EOF) status = SS$_NORMAL;
   sys$close (&FileFab, 0, 0);

   if (VMSnok (status))
   {
      if (!ListStarted)
      {
         ListStarted = true;
         fputs ("<ol>\n", stdout);
      }
      fprintf (stdout,
"<li><span class=\"hit\">ERROR reading file (%%X%08.08X) \
%s <!-- %s --></span>\n",
         status, FileNamePath, ResFileName);
   }

   if (HitCount)
   {
      if (!DocumentOnly) fputs ("</ol>\n", stdout);
      fflush (stdout);
      FileHitCount++;
   }
if (Debug) fprintf (stdout, "+++++4 %d\n", TagCharCount);
   if (TagCharCount & 1)
   {
      /* must have encountered an opening '<' without a closing '>' */
      if (!ListStarted)
      {
         ListStarted = true;
         fputs ("<ol>\n", stdout);
      }
      fprintf (stdout,
"<li><span class=\"hit\">HTML unbalanced &lt;&gt; in \
<a%s href=\"%s\">%s</a> <!-- %s --></span>\n",
               TargetPtr, FileNamePath, FileNamePath, ResFileName);
   }

   RecordCount += RecordNumber;

   return (status);
}

/*****************************************************************************/
/*
Search each record of what is presumed to be a plain-text file for the 
supplied string.  When the first hit occurs output an HTML anchor for 
retrieving the entire file.  For each hit output an HTML anchor to extract a 
specified range of record (lines) from the file.  This search only checks for 
at least one match in a line before considering it a hit.
*/

SearchTextFile ()

{
   int  idx,
        status,
        LastSectionRecordNumber = 1,
        RecordNumber = 0,
        ByteCount,
        HitCount = 0,
        MatchedLength;
   unsigned long  UnixTime;
   char  *cptr, *rptr, *sptr, *tptr,
         *CaseSensitivePtr = "",
         *FileNamePathPtr;
   char  FileSize [32],
         LastModifiedTime [64],
         Record [MAX_RECORD_SIZE+1],
         String [MAX_RECORD_SIZE*2];
   struct tm  *UnixTmPtr;
   struct FAB  FileFab;
#ifdef ODS_EXTENDED
   struct NAML  FileNaml;
#endif /* ODS_EXTENDED */
   struct RAB  FileRab;
   struct XABDAT  FileXabDat;
   struct XABFHC  FileXabFhc;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "SearchTextFile() |%s|%s|\n", ResFileName, SearchString);

   FileFab = cc$rms_fab;
   FileFab.fab$b_fac = FAB$M_GET;
   FileFab.fab$b_shr = FAB$M_SHRGET | FAB$M_SHRPUT;
   FileFab.fab$l_xab = &FileXabDat;
   FileXabDat = cc$rms_xabdat;
   FileXabDat.xab$l_nxt = &FileXabFhc;
   FileXabFhc = cc$rms_xabfhc;

#ifdef ODS_EXTENDED
   if (OdsExtended)
   {
      FileFab.fab$l_fna = (char*)-1;
      FileFab.fab$b_fns = 0;
      FileFab.fab$l_nam = (struct namdef*)&FileNaml;

      ENAMEL_RMS_NAML(FileNaml)
      FileNaml.naml$l_long_filename = ResFileName;
      FileNaml.naml$l_long_filename_size = ResFileNameLength;
   }
   else
#endif /* ODS_EXTENDED */
   {
      FileFab.fab$l_fna = ResFileName;  
      FileFab.fab$b_fns = ResFileNameLength;
   }

   status = sys$open (&FileFab, 0, 0);
   if (Debug) fprintf (stdout, "sys$open() %%X%08.08X\n", status);

   if (VMSnok (status))
   {
      if (status == RMS$_PRV) return (status);
      if (!ListStarted)
      {
         ListStarted = true;
         fputs ("<ol>\n", stdout);
      }
      fprintf (stdout,
"<li><span class=\"hit\">ERROR opening file (%%X%08.08X) \
%s <!-- %s --></span>\n",
         status, FileNamePath, ResFileName);
      return (SS$_NORMAL);
   }

   FileRab = cc$rms_rab;
   FileRab.rab$l_fab = &FileFab;
   /* 2 buffers and read ahead performance option */
   FileRab.rab$b_mbf = 2;
   FileRab.rab$l_rop = RAB$M_RAH;
   FileRab.rab$l_ubf = Record;
   FileRab.rab$w_usz = sizeof(Record)-1;

   if (VMSnok (status = sys$connect (&FileRab, 0, 0)))
   {
      if (Debug) fprintf (stdout, "sys$connect() %%X%08.08X\n", status);
      sys$close (&FileFab, 0, 0);
      return (status);
   }

   if (CaseSensitive) CaseSensitivePtr = "&case=yes";

   /**********************/
   /* search all records */
   /**********************/

   while (VMSok (status = sys$get (&FileRab, 0, 0)))
   {
      RecordNumber++;
      Record[FileRab.rab$w_rsz] = '\0';

      /* terminate on any carriage control that may be in the record */
      for (rptr = Record; *rptr && *rptr != '\r' && *rptr != '\n'; rptr++);
      *rptr = '\0';

      /* if necessary generate document name from first non-blank line */
      if (!DocumentName[0]) GenerateDocumentName (Record, DocumentName);

      if (*(rptr = Record))
         while (*rptr && isspace(*rptr)) rptr++;
      if (!*rptr)
      {
         /* the line contained none, or only white-space characters */
         LastSectionRecordNumber = RecordNumber + 1;
         continue;
      }

      tptr = SearchTextString (Record, SearchString,
                               CaseSensitive, true,
                               &MatchedLength);
      if (tptr != NULL)
      {
         /********/
         /* hit! */
         /********/

         if (Debug) fprintf (stdout, "Hit |%s|\n", tptr);

         TotalHitCount++;

#ifdef ODS_EXTENDED
         if (OdsExtended)
            FileNamePathPtr =
               (char*)CgiLibUrlEncodeFileName (FileNamePath, NULL, 0,
                                               FormatLikeVms, !FormatLikeVms);
         else
#endif /* ODS_EXTENDED */
            FileNamePathPtr = FileNamePath;

         if (!HitCount++)
         {
            /*************************************/
            /* first hit, entire document anchor */
            /*************************************/

            UnixTime = decc$fix_time (&FileXabDat.xab$q_rdt);
            UnixTmPtr = localtime (&UnixTime);
            if (!strftime (LastModifiedTime, sizeof(LastModifiedTime),
                           LastModifiedTimeFormatPtr, UnixTmPtr))
               strcpy (LastModifiedTime, "**ERROR**");

            /* true for non-VAR record formats, almost true for those :^) */
            if (FileXabFhc.xab$l_ebk)
               ByteCount = ((FileXabFhc.xab$l_ebk - 1) * 512) +
                           FileXabFhc.xab$w_ffb;
            else
               ByteCount = FileXabFhc.xab$w_ffb;
            if (ByteCount < 1000)
               sprintf (FileSize, "%d bytes", ByteCount);
            else
            if (ByteCount % 1000 < 500)
               sprintf (FileSize, "%d kB", ByteCount/1000);
            else
               sprintf (FileSize, "%d kB", (ByteCount/1000)+1);

            if (!ListStarted)
            {
               ListStarted = true;
               fputs ("<ol>\n", stdout);
            }

            fprintf (stdout,
"<li><span class=\"hit\">\
<a%s href=\"%s%s?highlight=%s%s%s%s&requery=%s\">%s</a> \
<span class=\"document\">TEXT &nbsp;%s &nbsp;%s</span></span>\n",
                     TargetPtr, ExtractScriptNamePtr, FileNamePathPtr,
                     UrlEncodedSearchString,
                     CaseSensitivePtr, FormText, FormHtml,
                     RequeryPtr,
                     DocumentName, LastModifiedTime, FileSize);

            if (DocumentOnly) break;

            fprintf (stdout, "<ol>\n");
         }

         /***********************/
         /* file extract anchor */
         /***********************/

         sptr = String;
         strcpy (sptr, "<li><span class=\"istext\">");
         sptr += 25;
         /* absorb leading white-space */
         for (cptr = Record; *cptr && (*cptr == ' ' || *cptr == '\t'); cptr++);
         /* copy the record up to the first character of the search string */
         sptr += CgiLibHtmlEscape (cptr, tptr-cptr, sptr, -1);

         /* add the HTML anchor */
         sptr += sprintf (sptr, "<a%s href=\"%s%s?highlight=%s&requery=%s",
                          TargetPtr, ExtractScriptNamePtr, FileNamePathPtr,
                          UrlEncodedSearchString, RequeryPtr);

         if (ExactNumberOfRecords)
            sptr += sprintf (sptr, "&start=%d&end=%d&exact=true%s%s\">",
                             RecordNumber,
                             RecordNumber+ExtractNumberOfRecords-1,
                             FormText, FormHtml);
         else
            sptr += sprintf (sptr, "&start=%d&end=%d%s%s\">",
                             LastSectionRecordNumber,
                             RecordNumber+ExtractNumberOfRecords-1,
                             FormText, FormHtml);

         /* matched string, highlighted within the anchor */
         sptr += CgiLibHtmlEscape (tptr, MatchedLength, sptr, -1);
         strcpy (sptr, "</a>");
         sptr += 4;

         /* rest of record after the matched search string */
         sptr += CgiLibHtmlEscape (tptr+MatchedLength, -1, sptr, -1);

         strcpy (sptr, "</span>\n");
         sptr += 8;

         if (Debug) fprintf (stdout, "String |%s|\n", String);
         fputs (String, stdout);
      }
   }

   /***************/
   /* end of file */
   /***************/

   if (status == RMS$_EOF) status = SS$_NORMAL;
   sys$close (&FileFab, 0, 0);

   if (VMSnok (status))
   {
      if (!ListStarted)
      {
         ListStarted = true;
         fputs ("<ol>\n", stdout);
      }

      fprintf (stdout,
"<li><span class=\"hit\">ERROR reading file (%%X%08.08X) \
%s <!-- %s --></span>\n",
         status, FileNamePath, ResFileName);
   }

   if (HitCount)
   {
      if (!DocumentOnly) fputs ("</ol>\n", stdout);
      fflush (stdout);
      FileHitCount++;
   }

   RecordCount += RecordNumber;

   return (status);
}

/*****************************************************************************/
/*
Escape HTML forbidden characters such as '<', '>' and '&' - except when the '&'
is the introducer for an HTML entity.  Leave the entity intact.
*/ 

int QueryHtmlEscape
(
char *string,
int length,
char *buf,
int size
)
{
   char  *aptr, *cptr, *czptr, *sptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "QueryHtmlEscape() |%s|\n", string);

   if (size < 0) size = 1024;  /* somewhat arbitrary */
   zptr = (sptr = buf) + size;
   if (length < 0)
      for (cptr = czptr = string; *czptr; czptr++);
   else
      czptr = (cptr = string) + length;
   for (; *cptr && cptr < czptr && sptr < zptr; cptr++)
   {
      if (*cptr == '&')
      {
         aptr = cptr + 1;
         if (*aptr == '#') aptr++;
         while (*aptr && !isspace(*aptr) && !ispunct(*aptr)) aptr++;
         if (*aptr == ';')
         {
            while (cptr < aptr && sptr < zptr) *sptr++ = *cptr++;
            if (sptr < zptr) *sptr++ = *cptr;
         }
         else
            for (aptr = "&amp;"; *aptr && sptr < zptr; *sptr++ = *aptr++);
      }
      else
      if (*cptr == '<')
         for (aptr = "&lt;"; *aptr && sptr < zptr; *sptr++ = *aptr++);
      else
      if (*cptr == '>')
         for (aptr = "&gt;"; *aptr && sptr < zptr; *sptr++ = *aptr++);
      else
         *sptr++ = *cptr;
   }
   *sptr = '\0';
   length = sptr - buf;
   if (sptr >= zptr && length > 10) memcpy (buf, "[OVERFLOW]", 10);

   return (length);
}

/*****************************************************************************/
/*
This function serves to generate a document name string (for use in the HTML
<TITLE> tag) for a plain-text file from the first line containing alpha-
numeric characters.  Copy the string pointed to by 'sptr' into the string 
pointed to by 'NamePtr', compressing white-space.
*/ 

GenerateDocumentName
(
char *sptr,
char *nptr
)
{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GenerateDocumentName() |%s|\n", sptr);

   /* skip leading non-alphanumerics */
   while (*sptr && !isalnum(*sptr)) sptr++;
   while (*sptr)
   {
      /* copy alphanumeric element */
      while (*sptr && !isspace(*sptr) && isalnum(*sptr)) *nptr++ = *sptr++;
      if (!*sptr) break;
      /* skip intervening/trailing non-alphanumerics */
      while (*sptr && !isalnum(*sptr)) sptr++;
      if (!*sptr) break;
      /* add a single space between alphanumeric elements */
      *nptr++ = ' ';
   }
   *nptr = '\0';
}

/*****************************************************************************/
/*
*/ 

QueryForm ()

{
   char  *cptr;
   char  EscapedPath [ODS_MAX_FILE_NAME_LENGTH+1];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "QueryForm()\n");

   FormatLikeVms = false;
   for (cptr = CgiPathInfoPtr; *cptr; cptr++);
   while (cptr > CgiPathInfoPtr && *cptr != '/')
   {
      if (*cptr == ';') FormatLikeVms = true;
      cptr--;
   }

   if (FormatLikeVms)
      CgiLibHtmlEscape (CgiPathTranslatedPtr, -1,
                        EscapedPath, sizeof(EscapedPath));
   else
      CgiLibHtmlEscape (CgiPathInfoPtr, -1,
                        EscapedPath, sizeof(EscapedPath));

   CgiLibResponseHeader (200, "text/html");

   fprintf (stdout,
"<!DOCTYPE html>\n\
<html>\n\
<head>\n\
<meta http-equiv=\"X-UA-Compatible\" content=\"IE=edge\">\n\
<meta name=\"generator\" content=\"%s\">\n\
<meta name=\"environment\" content=\"%s\">\n\
<title>Search %s</title>\n\
<style type=\"text/css\">\n\
%s\
</style>\n\
%s\
</head>\n\
<body>\n",
      SoftwareID,
      CgiEnvironmentPtr,
      EscapedPath,
      DefaultStyle,
      StyleSheetPtr);

   fprintf (stdout, "<div class=\"header\">Search %s</div>\n", EscapedPath);

   fprintf (stdout,
"<div class=\"query\">\n\
<form action=\"%s%s\">\n\
<input type=\"submit\" value=\"Search for:\">&nbsp;\n\
<input type=\"text\" name=\"search\" size=\"20\">&nbsp;\n\
<input type=\"reset\" value=\"reset\">\n\
<p>Case sensitive?&nbsp;\n\
<input type=\"radio\" name=\"case\" value=\"N\" checked>No\n\
<input type=\"radio\" name=\"case\" value=\"Y\">Yes\n\
<br>Treat HTML as if plain-text?&nbsp;\n\
<input type=\"radio\" name=\"plain\" value=\"N\" checked>No\n\
<input type=\"radio\" name=\"plain\" value=\"Y\">Yes\n\
<br>Open hit in new page?&nbsp;\n\
<input type=\"radio\" name=\"target\" value=\"_self\" checked>No\n\
<input type=\"radio\" name=\"target\" value=\"_blank\">Yes\n\
<br>Extract this number of lines around a plain-text &quot;hit&quot;:&nbsp;\n\
<input type=\"text\" name=\"extract\" value=\"25\" size=\"3\" maxlength=\"3\">\n\
</form>\n\
</div>\n",
      CgiScriptNamePtr, CgiPathInfoPtr);

   fprintf (stdout,
"<div class=\"footer\">\n\
<div class=\"rfoot\">\n\
<span class=\"about\"><a target=\"_blank\" href=\"%s\">about</a></span>\n\
</div>\n\
</div>\n",
            AboutPathPtr);

   fprintf (stdout, "</body>\n</html>\n");
}

/****************************************************************************/
/*
Return an integer reflecting the major and minor version of VMS (e.g. 60, 61,
62, 70, 71, 72, etc.)
*/ 

#ifdef ODS_EXTENDED

int GetVmsVersion ()

{
   static char  SyiVersion [16];

   static struct {
      short int  buf_len;
      short int  item;
      void  *buf_addr;
      unsigned short  *ret_len;
   }
   SyiItems [] =
   {
      { 8, SYI$_VERSION, &SyiVersion, 0 },
      { 0,0,0,0 }
   };

   int  status,
        version;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "GetVmsVersion()\n");

   if (VMSnok (status = sys$getsyiw (0, 0, 0, &SyiItems, 0, 0, 0)))
      exit (status);
   SyiVersion[8] = '\0';
   version = ((SyiVersion[1]-48) * 10) + (SyiVersion[3]-48);
   if (Debug) fprintf (stdout, "|%s| %d\n", SyiVersion, version);
   return (version);
}

#endif /* ODS_EXTENDED */

/****************************************************************************/
/*
Does a case-insensitive, character-by-character string compare and returns
true if two strings are the same, or false if not.
*/

BOOL strsame
(
char *sptr1,
char *sptr2,
int  count
)
{
   /*********/
   /* begin */
   /*********/

   if (count > 0)
      return (strncasecmp (sptr1, sptr2, count) == 0);
   else
      return (strcasecmp (sptr1, sptr2) == 0);
}

/*****************************************************************************/

